package blatt05;

import blatt05.alphabet.Alphabet;
import blatt05.alphabet.SmallIntegerAlphabet;
import blatt05.alphabet.DNAAlphabet;
import java.io.File;
import java.util.Arrays;
import java.util.HashMap;

/**
 * Implementation of a q-gram profile backed by a dense array.
 * @author Nils.Hoffmann@CeBiTec.Uni-Bielefeld.DE
 */
public class DenseQGramProfile implements QGramProfile{
	
	private int q = 0;
	
	private int alphabetsize = 0;

        private int[] profile = null;

        private Alphabet alph = null;

        private boolean rising = false;
	
	public DenseQGramProfile(int alphabetsize, int q, String word, boolean rising, Alphabet alph) {
		this(alphabetsize, q, toIntArray(word,alph), rising, alph);
	}

        public DenseQGramProfile(int alphabetsize, int q, int[] word, boolean rising, Alphabet alph) {
            if(q<1) {
                throw new IllegalArgumentException("q must be larger than 0!");
            }
            if(alphabetsize!=alph.size()) {
                System.err.println(alphabetsize+"!="+alph.size());
                throw new IllegalArgumentException("Alphabetsize and codeTable size must be of the same value!");
            }
		this.q = q;
		this.alphabetsize = alphabetsize;
                this.rising = rising;
                this.profile = new int[(int)Math.pow(this.alphabetsize,q)];
                this.alph = alph;
                
                System.out.println("Profile size: "+this.profile.length);
                if(rising){
                    getProfileRising(word,true,this.profile);
                }else{
                    getProfileFalling(word,true,this.profile);
                }
	}
	
	private int[] getProfileFalling(int[] word, boolean add, int[] profile) {
                int[] pp = profile==null?new int[(int)Math.pow(this.alphabetsize,q)]:profile;
		int rank = -1;
		for(int i=0;i<word.length-this.q+1;i++) {
			rank = rankFalling(i, word, rank,this.q);
			//System.out.println("Falling Rank for "+Arrays.toString(Arrays.copyOfRange(word, i, i+this.q))+" = "+rank);
			if(add){
                            pp[rank]++;
                        }else{
                            pp[rank]--;
                        }
		}
                return pp;
	}
	
	private int[] getProfileRising(int[] word, boolean add, int[] profile) {
                int[] pp = profile==null?new int[(int)Math.pow(this.alphabetsize,q)]:profile;
		int rank = -1;
		for(int i=0;i<word.length-this.q+1;i++) {
			rank = rankRising(i, word, rank,this.q);
			//System.out.println("Rising Rank for "+Arrays.toString(Arrays.copyOfRange(word, i, i+this.q))+" = "+rank);
			if(add){
                            pp[rank]++;
                        }else{
                            pp[rank]--;
                        }
		}
		return pp;
	}
	
	private int rankFalling(int j, int[] word, int pRank, int q) {
		int rank = 0;
		if(pRank!=-1) {
			int pos = (int)Math.pow(alphabetsize,q-1);
			rank = (pRank%(pos))*alphabetsize + word[j+q-1];
                        //System.out.println("Code for position "+(q-1)+": previous Rank=" +pRank+" pos(sigma^{q-1})="+pos+" factor(previousRank%pos)="+(pRank%pos)+" * "+alphabetsize+" + "+word[j+q-1]+" => "+rank);
		}else{
			for(int i=0;i<q;i++) {
                            int v = (int)(word[j+i]*(Math.pow(alphabetsize,(q-i-1))));
                            //System.out.println("Code for position "+(q-i-1)+": factor="+word[j+i]+" base="+alphabetsize+" exponent="+(q-i-1)+" => "+word[j+i]*(Math.pow(alphabetsize,(q-i-1))));
                            rank+=v;
			}
		}
		return rank;
	}
	
	private int rankRising(int j, int [] word, int pRank, int q) {
		int rank = 0;
		if(pRank!=-1) {
                        int pos = (int)Math.pow(alphabetsize,q-1)*word[j+q-1];
			rank = ((int)pRank/alphabetsize) + pos;
                        //System.out.println("Code for position "+(q-1)+": previous Rank=" +pRank+" pos(sigma^{q-1}*word[j+q-1])="+pos+" factor(previousRank/sigma)="+(pRank/alphabetsize)+" + "+pos+" => "+rank);
		}else{
			for(int i=0;i<q;i++) {
                            int v = (int)(word[j+i]*(Math.pow(alphabetsize,(i))));
                            //System.out.println("Code for position "+i+": factor="+word[j+i]+" base="+alphabetsize+" exponent="+i+" => "+word[j+i]*(Math.pow(alphabetsize,(i))));
                            rank+=v;
			}
		}
			
		return rank;
	}
	
	private int[] unrankFalling(int rank, int q) {
		int rest = rank;
		System.out.println("Unrank falling on: "+rest);
		int [] word = new int[q];
		for(int i=0;i<q;i++) {
			int pos = (int)Math.pow(this.alphabetsize,q-i-1);
			int div = rest/pos;
			word[q-i-1] = div;
			rest = rest - div*pos;
			System.out.println(" word="+word[q-i-1]+" rest="+rest);
		}
                System.out.println("Reverse rank:"+rankFalling(0,word,-1, q));
		return word;
	}
	
	private int[] unrankRising(int rank, int q) {
		int rest = rank;
		System.out.println("Unrank rising on: "+rest);
		int [] word = new int[q];
		for(int i=0;i<q;i++) {
			int pos = (int)Math.pow(this.alphabetsize,q-i-1);
			int div = rest/pos;
			word[i] = div;
			rest = rest - div*pos;
			System.out.println(" word="+word[i]+" rest="+rest);
		}
		return word;
        }

        public static int[] toIntArray(String a,Alphabet alph) {
            int[] vals = new int[a.length()];
            for(int i = 0; i<vals.length; i++) {
                vals[i] = alph.toCode(a.charAt(i)+"");
            }
            return vals;
        }

        public static String toString(int[] v, Alphabet alph) {
            char[] c = new char[v.length];
            for(int i = 0;i<c.length;i++) {
                c[i] = alph.toSymbol(v[i]).charAt(0);
            }
            String s= new String(c);
            return s;
        }

    @Override
        public String toString() {
            StringBuilder sb = new StringBuilder();
            for(int i = 0; i<this.profile.length; i++) {
                int cnt = getOccurrenceFor(i);
                if(cnt!=0) {
                    sb.append((this.rising?toString(unrankRising(i, this.q),this.alph):toString(unrankFalling(i, this.q),this.alph))+": "+cnt+"\n");
                }
            }
            return sb.toString();
        }
    
	public static void main(String[] args) {
//            int q = 4;
//            int sigma = 8;
//            Alphabet a = new SmallIntegerAlphabet();
//            QGramProfile qgp = new DenseQGramProfile(a.size(),q,new int[]{3,2,6,7,1},true,a);
//            int[] val = qgp.unrankRising(3267, 4);
//            System.out.println(Arrays.toString(val));
//
//            qgp = new DenseQGramProfile(sigma,q,new int[]{3,2,6,7,1},false,a);
//            val = qgp.unrankFalling(3267, 4);
//            System.out.println(Arrays.toString(val));
            FASTAReader fr = new FASTAReader();
            Alphabet b = new DNAAlphabet();
            HashMap<String,String> hm = fr.read(new File("xanthomonas.fasta"));
//            for(int i=1;i<100;i++) {
//                for(String key:hm.keySet()) {
//                    System.out.println("q="+i);
//                    QGramProfile qgp = new SparseQGramProfile(b.size(), i, hm.get(key), true, b);
//                }
//            }
            for(int i = 1;i<100;i++) {
                for(String key:hm.keySet()) {
                    System.out.println("q="+i);
                    QGramProfile qgp = new DenseQGramProfile(b.size(),i,hm.get(key),true,b);
                        //System.out.println(qgp.toString());
                }
            }

//		int[] values = new int[]{2,1,7,5,3};
//                System.out.println("Size of profile: "+Math.pow(sigma,q));
//		int[] pf = qgp.getProfileFalling(values);
//		for(int i=0;i<pf.length;i++) {
//			if(pf[i]>0) {
//				System.out.println(i+" = "+pf[i]);
//			}
//		}
//		int[] pr = qgp.getProfileRising(values);
//		for(int i=0;i<pr.length;i++) {
//			if(pr[i]>0) {
//				System.out.println(i+" = "+pr[i]);
//			}
//		}
//		System.out.println(Arrays.toString(qgp.unrankFalling(3018,q)));
//		System.out.println(Arrays.toString(qgp.unrankFalling(1913,q)));
//		System.out.println(Arrays.toString(qgp.unrankRising(1149,q)));
//		System.out.println(Arrays.toString(qgp.unrankRising(1003,q)));
	}

    public int getOccurrenceFor(int code) {
       return this.profile[code];
    }

    public int getOccurrenceFor(String qgram) {
        return getOccurrenceFor(this.rising?rankRising(0, toIntArray(qgram,this.alph), 0, this.q):rankFalling(0,toIntArray(qgram,this.alph),0,this.q));
    }

    public void subtractFromProfile(String b) {
        if(this.rising) {
            getProfileRising(toIntArray(b,this.alph),false,this.profile);
        }else{
            getProfileFalling(toIntArray(b,this.alph),false,this.profile);
        }
    }

    public int size() {
        return this.profile==null?-1:this.profile.length;
    }

    public int getRank(String word) {
        if(this.rising) {
            return rankRising(0, toIntArray(word, this.alph), -1, this.q);
        }else{
            return rankFalling(0, toIntArray(word, this.alph), -1, this.q);
        }
    }

    public String getWord(int rank) {
        if(this.rising) {
            return toString(unrankRising(rank, this.q),this.alph);
        }else{
            return toString(unrankFalling(rank, this.q),this.alph);
        }
    }
	
}
