package blatt05;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;

/**
 * This class provides an implementation of a simple
 * parser for the FASTA format.
 *
 * @author Nils.Hoffmann@CeBiTec.Uni-Bielefeld.DE
 */
public class FASTAReader {

	/**
	 * Reads FASTA sequences from a given File object.
	 * Multiple sequences are supported.
	 * @param f The input file to read from.
	 * @returns a HashMap object holding sequence ids as keys and sequences as objects
	 */
    public HashMap<String,String> read(File f) {
        HashMap<String,String> sequences = new HashMap<String,String>();
        try {
			BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(f)));
			String line = "";
			StringBuilder blockBuilder = new StringBuilder(128);
			String sname = "";
			try {
				//Read one line of input
				while((line=br.readLine())!=null) {
					//Init new Sequence after comment line
					if(line.startsWith(">")) {
						//postprocess if we have attached all lines
						if(blockBuilder.length()>0) {
							sequences.put(sname,blockBuilder.toString());
                                                        //reset blockBuilder
							blockBuilder = new StringBuilder(128);
						}
						sname = line.substring(line.indexOf(">")+1);
					} else {
						//process line and add to return list
						line.trim();
						if(!line.isEmpty()) {
							blockBuilder.append(line);
						}
					}
				}
				//catch the last block in a file
				if(blockBuilder.length()>0) {
					sequences.put(sname,blockBuilder.toString());
				}
			} catch (IOException e) {
				System.err.println(e.getLocalizedMessage());
			}
		} catch (FileNotFoundException e) {
			System.err.println("File not found: "+f.getAbsolutePath());
			System.err.println(e.getLocalizedMessage());
		}
		return sequences;
    }
}
