package blatt05;

import blatt05.alphabet.Alphabet;
import java.util.Arrays;
import java.util.HashMap;

/**
 * Implementation of a q-gram profile backed by a HashMap.
 * @author Nils.Hoffmann@CeBiTec.Uni-Bielefeld.DE
 */
public class SparseQGramProfile implements QGramProfile{
        private int q = 0;

	private int alphabetsize = 0;

        private HashMap<Integer,Integer> profile = null;

        private int virtualSize = -1;

        private Alphabet alph = null;

        private boolean rising = false;

	public SparseQGramProfile(int alphabetsize, int q, String word, boolean rising, Alphabet alph) {
		this(alphabetsize, q, toIntArray(word,alph), rising, alph);
	}

        public SparseQGramProfile(int alphabetsize, int q, int[] word, boolean rising, Alphabet alph) {
            if(q<1) {
                throw new IllegalArgumentException("q must be larger than 0!");
            }
            if(alphabetsize!=alph.size()) {
                System.err.println(alphabetsize+"!="+alph.size());
                throw new IllegalArgumentException("Alphabetsize and codeTable size must be of the same value!");
            }
		this.q = q;
		this.alphabetsize = alphabetsize;
                this.rising = rising;
                this.profile = new HashMap<Integer,Integer>();
                this.alph = alph;
                this.virtualSize = (int)Math.pow(this.alphabetsize,q);
                System.out.println("Profile size: "+this.virtualSize);
                if(rising){
                    getProfileRising(word,true,this.profile);
                }else{
                    getProfileFalling(word,true,this.profile);
                }
	}

	public HashMap<Integer,Integer> getProfileRising(int[] word, boolean add, HashMap<Integer,Integer> profile) {
		int rank = -1;
		for(int i=0;i<word.length-this.q+1;i++) {
			rank = rankRising(i, word, rank,this.q);
			//System.out.println("Rising Rank for "+Arrays.toString(Arrays.copyOfRange(word, i, i+this.q))+" = "+rank);
                        Integer key = Integer.valueOf(rank);
			if(add){
                            if(profile.containsKey(key)){
                                profile.put(key, Integer.valueOf(profile.get(key)+1));
                            }else{
                                profile.put(key,Integer.valueOf(1));
                            }
                        }else{
                            if(profile.containsKey(key)){
                                profile.put(key, Integer.valueOf(profile.get(key)-1));
                            }else{
                                profile.put(key,Integer.valueOf(-1));
                            }
                        }
		}
                return profile;
	}

	public HashMap<Integer,Integer> getProfileFalling(int[] word, boolean add, HashMap<Integer,Integer> profile) {
		int rank = -1;
		for(int i=0;i<word.length-this.q+1;i++) {
			rank = rankFalling(i, word, rank,this.q);
                        Integer key = Integer.valueOf(rank);
			//System.out.println("Falling Rank for "+Arrays.toString(Arrays.copyOfRange(word, i, i+this.q))+" = "+rank);
			if(add){
                            if(profile.containsKey(key)){
                                profile.put(key, Integer.valueOf(profile.get(key)+1));
                            }else{
                                profile.put(key,Integer.valueOf(1));
                            }
                        }else{
                            if(profile.containsKey(key)){
                                profile.put(key, Integer.valueOf(profile.get(key)-1));
                            }else{
                                profile.put(key,Integer.valueOf(-1));
                            }
                        }
		}
		return profile;
	}

	public int rankFalling(int j, int[] word, int pRank, int q) {
		int rank = 0;
		if(pRank!=-1) {
			int pos = (int)Math.pow(alphabetsize,q-1);
			rank = (pRank%(pos))*alphabetsize + word[j+q-1];
		}else{
			for(int i=0;i<q;i++) {
				rank+=(word[j+i]*(Math.pow(alphabetsize,(q-i-1))));
			}
		}
		return rank;
	}

	public int rankRising(int j, int [] word, int pRank, int q) {
		int rank = 0;
		if(pRank!=-1) {
			rank = ((int)pRank/alphabetsize) + (int)Math.pow(alphabetsize,q-1)*word[j+q-1];
		}else{
			for(int i=0;i<q;i++) {
				rank+=(word[j+i]*(Math.pow(alphabetsize,(i))));
			}
		}

		return rank;
	}

	public int[] unrankFalling(int rank, int q) {
		int rest = rank;
		//System.out.println("Rank: "+rest);
		int [] word = new int[q];
		for(int i=0;i<q;i++) {
			int pos = (int)Math.pow(this.alphabetsize,q-i-1);
			int div = rest/pos;
			word[q-i-1] = div;
			rest = rest - div*pos;
			//System.out.println(" word="+word[q-i-1]+" rest="+rest);

		}
		return word;
	}

	public int[] unrankRising(int rank, int q) {
		int rest = rank;
		System.out.println("Unrank rising on: "+rest);
		int [] word = new int[q];
		for(int i=0;i<q;i++) {
			int pos = (int)Math.pow(this.alphabetsize,q-i-1);
			int div = rest/pos;
			word[i] = div;
			rest = rest - div*pos;
			System.out.println(" word="+word[i]+" rest="+rest);
		}
		return word;
        }

        public static int[] toIntArray(String a,Alphabet alph) {
            int[] vals = new int[a.length()];
            for(int i = 0; i<vals.length; i++) {
                vals[i] = alph.toCode(a.charAt(i)+"");
            }
            return vals;
        }

        public static String toString(int[] v, Alphabet alph) {
            char[] c = new char[v.length];
            for(int i = 0;i<c.length;i++) {
                c[i] = alph.toSymbol(v[i]).charAt(0);
            }
            String s= new String(c);
            return s;
        }

    @Override
        public String toString() {
            StringBuilder sb = new StringBuilder();
            for(Integer key:this.profile.keySet()) {
                int cnt = getOccurrenceFor(key.intValue());
                if(cnt!=0) {
                    sb.append((this.rising?toString(unrankRising(key.intValue(), this.q),this.alph):toString(unrankFalling(key.intValue(), this.q),this.alph))+": "+cnt+"\n");
                }
            }
            return sb.toString();
        }

	

    public int getOccurrenceFor(int code) {
       return this.profile.get(Integer.valueOf(code));
    }

    public int getOccurrenceFor(String qgram) {
        return getOccurrenceFor(this.rising?rankRising(0, toIntArray(qgram,this.alph), 0, this.q):rankFalling(0,toIntArray(qgram,this.alph),0,this.q));
    }

    public void subtractFromProfile(String b) {
        if(this.rising) {
            getProfileRising(toIntArray(b,this.alph),false,this.profile);
        }else{
            getProfileFalling(toIntArray(b,this.alph),false,this.profile);
        }
    }

    public int size() {
        return this.virtualSize;
    }

    public int getRank(String word) {
        if(this.rising) {
            return rankRising(0, toIntArray(word, this.alph), -1, this.q);
        }else{
            return rankFalling(0, toIntArray(word, this.alph), -1, this.q);
        }
    }

    public String getWord(int rank) {
        if(this.rising) {
            return toString(unrankRising(rank, this.q),this.alph);
        }else{
            return toString(unrankFalling(rank, this.q),this.alph);
        }
    }
}
